
//
// JPHP - JIT optimizer for the PHP language
// Copyright (C) 2001-2003 Indrek Mandre <indrek(at)mare.ee>
//
// This source code is made available free and without charge subject
// to the terms of the QPL as detailed in bundled LICENSE.QPL file.
//
// It may be only used for non-commercial purpose.
//

#include "jphp.hh"
#include "optimize.hh"

Cell::Cell ()
{
  memset (&node, 0, sizeof (ex_node));
  SET_UNUSED (node.result);
  SET_UNUSED (node.op1);
  SET_UNUSED (node.op2);
  node.result.op_type = JOP_UNUSED;
  node.op1.op_type = JOP_UNUSED;
  node.op2.op_type = JOP_UNUSED;
  SET_UNUSED (op1);
  SET_UNUSED (op2);
  node.opcode = ZEND_NOP;
  next_jmp = NULL;
  j2_next = NULL;
  order_id = 0;
}

static inline void op_transform (jnode&to, const znode& from)
{
  to.op_type = from.op_type;
  to.EA_type = from.u.EA.type;
  to.u.var = from.u.var;
  switch (to.op_type)
    {
      case IS_VAR:
        to.op_type = JOP_VAR;
        break;
      case IS_TMP_VAR:
        to.op_type = JOP_TMP;
        break;
      case IS_CONST:
        to.op_type = JOP_CONST;
        break;
      case IS_UNUSED:
      default:
        to.op_type = JOP_UNUSED;
        break;
    }
}

Cell::Cell (zend_op *op)
{
  memset (&node, 0, sizeof (ex_node));
  node.opcode = op->opcode;

  op_transform (node.result, op->result);
  op_transform (node.op1, op->op1);
  op_transform (node.op2, op->op2);

  op1 = op->op1;
  op2 = op->op2;
  extended_value = node.jex.extended_value = op->extended_value;
  node.lineno = op->lineno;

  next_jmp = NULL;
  j2_next = NULL;

  order_id = 0;

  node.jext_type = JEXT_UNUSED;
}

void Cell::alt_jmp_to (Cell *to)
{
  if ( next_jmp )
    {
      jump_to.erase (next_jmp);
      next_jmp->jump_from.erase (this);
    }
  next_jmp = to;
  jump_to.insert (to);
  next_jmp->jump_from.insert (this);
}

void Cell::replace (Cell *c)
{
  set<Cell *>::iterator it = jump_from.begin();
  while ( it != jump_from.end() )
    {
      Cell &j = **it;

      if ( j.next_jmp == this )
        {
          j.next_jmp = c;
          j.jump_to.erase (this);
          j.jump_to.insert (c);
          c->jump_from.insert (&j);
        }

      if ( j.j2_next == this )
        {
          j.j2_next = c;
          j.jump_to.erase (this);
          j.jump_to.insert (c);
          c->jump_from.insert (&j);
        }
      it++;
    }

  jump_from.clear ();
  jump_to.clear ();
}

bool Cell::resolve_vars (Cell *parent)
{
  return true;
}


