
//
// JPHP - JIT optimizer for the PHP language
// Copyright (C) 2001-2003 Indrek Mandre <indrek(at)mare.ee>
//
// This source code is made available free and without charge subject
// to the terms of the QPL as detailed in bundled LICENSE.QPL file.
//
// It may be only used for non-commercial purpose.
//

#include <cstdio>
#include <cstdlib>

#include <sys/types.h>
#include <sys/stat.h>

#include <string>
#include <vector>
#include <set>

#include "jphp.hh"

#include "data.hh"

void decode_oparray (zend_op_array* array, Datainput& data, jphp_header *head)
{
  array->static_variables = data.recover_hashtable_of_zval (array->static_variables, head->consts);
  data.recover (array->arg_types);
  data.recover (array->function_name);
  data.recover (array->brk_cont_array);
  data.recover (array->filename);

  ex_node *ex = reinterpret_cast<ex_node *>(array->opcodes);

  for ( size_t j = 0; j < array->size; j++ )
    {
      ex_node &node = ex[j];

      node.jump = jphp_jump[node.opcode];

      if ( node.op1.op_type == JOP_CONST || node.op1.op_type == JOP_FETCH )
          node.op1.u.cnst = &head->consts[node.op1.u.var].val;

      if ( node.op2.op_type == JOP_CONST || node.op2.op_type == JOP_FETCH )
          node.op2.u.cnst = &head->consts[node.op2.u.var].val;

      if ( node.jext_type == JEXT_ALT_JMP )
          node.jex.alt_jump.next = &ex[reinterpret_cast<size_t>(node.jex.alt_jump.next)];
      else if ( node.jext_type == JEXT_FENTRY )
          node.jex.fentry = &head->functions[reinterpret_cast<size_t>(node.jex.fentry)];
      else if ( node.jext_type == JEXT_CENTRY )
          node.jex.centry = &head->classes[reinterpret_cast<size_t>(node.jex.centry)];
      else if ( node.jext_type == JEXT_HEADER )
          node.jex.header = head;
      else if ( node.jext_type == JEXT_BOP )
        {
          switch (reinterpret_cast<int>(node.jex.bop.op))
            {
              case ZEND_ADD:
              case ZEND_ASSIGN_ADD:
                node.jex.bop.op = add_function;
                break;
              case ZEND_SUB:
              case ZEND_ASSIGN_SUB:
                node.jex.bop.op = sub_function;
                break;
              case ZEND_MUL:
              case ZEND_ASSIGN_MUL:
                node.jex.bop.op = mul_function;
                break;
              case ZEND_DIV:
              case ZEND_ASSIGN_DIV:
                node.jex.bop.op = div_function;
                break;
              case ZEND_MOD:
              case ZEND_ASSIGN_MOD:
                node.jex.bop.op = mod_function;
                break;
              case ZEND_SL:
              case ZEND_ASSIGN_SL:
                node.jex.bop.op = shift_left_function;
                break;
              case ZEND_SR:
              case ZEND_ASSIGN_SR:
                node.jex.bop.op = shift_right_function;
                break;
              case ZEND_CONCAT:
              case ZEND_ASSIGN_CONCAT:
                node.jex.bop.op = concat_function;
                break;
              case ZEND_BW_OR:
              case ZEND_ASSIGN_BW_OR:
                node.jex.bop.op = bitwise_or_function;
                break;
              case ZEND_BW_AND:
              case ZEND_ASSIGN_BW_AND:
                node.jex.bop.op = bitwise_and_function;
                break;
              case ZEND_BW_XOR:
              case ZEND_ASSIGN_BW_XOR:
                node.jex.bop.op = bitwise_xor_function;
                break;
              case ZEND_IS_IDENTICAL:
                node.jex.bop.op = is_identical_function;
                break;
              case ZEND_IS_NOT_IDENTICAL:
                node.jex.bop.op = is_not_identical_function;
                break;
              case ZEND_IS_EQUAL:
                node.jex.bop.op = is_equal_function;
                break;
              case ZEND_IS_NOT_EQUAL:
                node.jex.bop.op = is_not_equal_function;
                break;
              case ZEND_IS_SMALLER:
                node.jex.bop.op = is_smaller_function;
                break;
              case ZEND_IS_SMALLER_OR_EQUAL:
                node.jex.bop.op = is_smaller_or_equal_function;
                break;
              case ZEND_BOOL_XOR:
                node.jex.bop.op = boolean_xor_function;
                break;
            }
        }

      if ( node.op2.op_type == JOP_JMP )
          node.op2.u.jmp_to = &ex[node.op2.u.opline_num];

    }
}

struct jphp_header * jphp_load (char *buf, size_t buf_len)
{
  struct jphp_header *head;
  head = reinterpret_cast<struct jphp_header*>(buf);

  buf += sizeof (struct jphp_header);

  head->functions = reinterpret_cast<struct jphp_fentry *>(buf);

  buf += sizeof (struct jphp_fentry) * head->function_count;

  for ( size_t i = 0; i < head->function_count; i++ )
    {
      head->functions[i].op_array = reinterpret_cast<zend_op_array *>(buf);
      buf += sizeof (zend_op_array);
      head->functions[i].op_array->opcodes = reinterpret_cast<zend_op *> (buf);
      buf += head->functions[i].op_array->size * sizeof (struct ex_node);
    }

  head->classes = reinterpret_cast<jphp_centry *>(buf);
  buf += sizeof (jphp_centry) * head->class_count;

  for ( size_t i = 0; i < head->class_count; i++ )
    {
      HashTable *ht = reinterpret_cast<HashTable *>(buf);
      buf += sizeof (HashTable);
      ht->arBuckets = reinterpret_cast<Bucket **>(buf);
      buf += sizeof (Bucket *) * ht->nTableSize;

      Bucket *prev = NULL;
      for ( size_t j = 0; j < ht->nNumOfElements; j++ )
        {
          Bucket *b = reinterpret_cast<Bucket *>(buf);
          buf += sizeof (Bucket) + b->nKeyLength - 1;
          zend_op_array *op_array = reinterpret_cast<zend_op_array *>(buf);
          buf += sizeof (zend_op_array);
          op_array->opcodes = reinterpret_cast<zend_op *> (buf);
          buf += op_array->size * sizeof (struct ex_node);

          b->pData = op_array;
          b->pListLast = prev;

          if ( !j )
              ht->pListHead = b;
          else
              prev->pListNext = b;

          size_t idx = b->h & ht->nTableMask;
          b->pNext = ht->arBuckets[idx];
          if ( ht->arBuckets[idx] ) ht->arBuckets[idx]->pLast = b;
          ht->arBuckets[idx] = b;

          prev = b;
        }

      ht->pListTail = prev;
      head->classes[i].function_table = ht;
    }

  size_t csize = *reinterpret_cast<size_t *>(buf);
  buf += sizeof (size_t);
  head->consts = reinterpret_cast<jphp_const *>(buf);
  head->const_count = csize;
  buf += sizeof (jphp_const) * csize;

  Datainput data(buf);

  // At first process all consts and recover their data
  for ( size_t i = 0; i < csize; i++ )
    {
      zval &val = head->consts[i].val;
      if ( val.type == IS_STRING || val.type == IS_CONSTANT ) data.recover (val.value.str.val);
      if ( val.type == IS_ARRAY || val.type == IS_CONSTANT_ARRAY )
          val.value.ht = data.recover_hashtable_of_zval (val.value.ht, head->consts);
    }

  data.recover (head->path);

  for ( size_t i = 0; i < head->function_count; i++ )
    {
      data.recover (head->functions[i].name);
      decode_oparray (head->functions[i].op_array, data, head);
    }

  for ( size_t i = 0; i < head->class_count; i++ )
    {
      jphp_centry& c = head->classes[i];
      data.recover (c.centry);
      data.recover (c.centry->name);  // c.centry must not be rerecovered
      data.recover (c.name);
      data.recover (c.parent);
      c.default_properties = data.recover_hashtable_of_zval (c.default_properties, head->consts);
      memcpy (&c.centry->function_table, c.function_table, sizeof (HashTable));
      memcpy (&c.centry->default_properties, c.default_properties, sizeof (HashTable));

      php_hash<zend_op_array> ht(c.function_table);
      ht.reset();
      while ( zend_op_array *arr = ht.each() )
          decode_oparray (arr, data, head);
    }

  return head;
}

void add_class_member (zend_op_array *op)
{
  static zend_uint large_refcount;
  large_refcount = 666666666;
  op->refcount = &large_refcount;

  if ( op->static_variables )
    {
      HashTable *ht = op->static_variables;
      zval *tmp;

      ALLOC_HASHTABLE_REL(op->static_variables);
      zend_hash_init(op->static_variables, 0, NULL, ZVAL_PTR_DTOR, 0);
      zend_hash_copy(op->static_variables, ht, (copy_ctor_func_t) zval_add_ref,
          (void *) &tmp, sizeof(zval *));
    }
}

void jphp_register_function (ex_node *node)
{
  switch (node->op1.u.register_type)
    {
      case ZEND_DECLARE_FUNCTION:
        {
          jphp_fentry& fentry = *node->jex.fentry;

          zend_function *function, *rfun;
          php_hash<zend_function> ft(EG(function_table));

          function = reinterpret_cast<zend_function *>(fentry.op_array);

          if ( !ft.quick_insert (fentry.name, fentry.name_len + 1, fentry.name_hcode, *function, rfun) ) {
            if ( ft.quick_find (fentry.name, fentry.name_len + 1, fentry.name_hcode, function)
              && function->type==ZEND_USER_FUNCTION
              && ((zend_op_array *) function)->last>0) {
              zend_error(E_ERROR, "Cannot redeclare %s() (previously declared in %s:%d)",
                    fentry.name,
                    ((zend_op_array *) function)->filename,
                    jphp_function_lineno (function));
            } else {
              zend_error(E_ERROR, "Cannot redeclare %s()", fentry.name);
            }
          } else {
            if ( rfun->op_array.static_variables )
              {
                HashTable *ht = rfun->op_array.static_variables;
                zval *tmp;

                ALLOC_HASHTABLE_REL(rfun->op_array.static_variables);
                zend_hash_init(rfun->op_array.static_variables, 0, NULL, ZVAL_PTR_DTOR, 0);
                zend_hash_copy(rfun->op_array.static_variables, ht, (copy_ctor_func_t) zval_add_ref, (void *) &tmp, sizeof(zval *));
              }

            static zend_uint large_refcount;
            large_refcount = 666666666;
            rfun->op_array.refcount = &large_refcount;
          }
          break;
        }
      case ZEND_DECLARE_CLASS:
      case ZEND_DECLARE_INHERITED_CLASS:
        {
          jphp_register_class (*node->jex.centry);
          break;
        }
    }
}

int jphp_function_lineno (zend_function *function)
{
  zend_op_array *oa = &function->op_array;
  if ( oa->type != ZEND_USER_FUNCTION || ! oa->opcodes ) return 0;
  if ( oa->opcodes[0].opcode == ZEND_JEXT_START )
    {
      ex_node *en = reinterpret_cast<ex_node *>(&oa->opcodes[0]);
      return en->lineno;
    }
  else
    {
      return oa->opcodes[0].lineno;
    }
}


void jphp_register_class (jphp_centry& centry)
{
  zend_class_entry& ce = *centry.centry;
  zend_class_entry *parent, *ret;

  php_hash<zend_class_entry> ct(EG(class_table));

  if ( centry.parent )
      if ( !ct.quick_inline_find (centry.parent, centry.parent_len + 1, centry.parent_hcode, parent) )
          zend_error(E_ERROR, "Class %s:  Cannot inherit from undefined class %s", centry.name, centry.parent);

  if ( !ct.quick_insert (centry.name, centry.name_len + 1, centry.name_hcode, ce, ret) )
      zend_error(E_ERROR, "Cannot redeclare class %s", centry.name);

  ret->refcount = static_cast<int *>(emalloc (sizeof (int)));
  *ret->refcount = 1;
  ret->name = estrndup (ret->name, ret->name_length);

  typedef void (*cast_t)(void*);

  zend_hash_init (&ret->function_table, centry.function_table->nTableSize, NULL, ZEND_FUNCTION_DTOR, 0);
  zend_function tmp_function;
  zend_hash_copy (&ret->function_table, centry.function_table, reinterpret_cast<void (*)(void*)>(add_class_member), &tmp_function,
      sizeof (zend_function));

  zend_hash_init (&ret->default_properties, centry.default_properties->nTableSize, NULL, ZVAL_PTR_DTOR, 0);
  zval *tmp_zval;
  zend_hash_copy (&ret->default_properties, centry.default_properties, reinterpret_cast<void (*)(void*)>(zval_add_ref), (void *)&tmp_zval,
      sizeof (zval *));

  if ( centry.parent )
      zend_do_inheritance(ret, parent); // all merged functins already have processed static vars, so don't worry
}


