
//
// JPHP - JIT optimizer for the PHP language
// Copyright (C) 2001-2003 Indrek Mandre <indrek(at)mare.ee>
//
// This source code is made available free and without charge subject
// to the terms of the QPL as detailed in bundled LICENSE.QPL file.
//
// It may be only used for non-commercial purpose.
//

#include "jphp.hh"

static char *opname[] = {
"ZEND_NOP",  // 0
"ZEND_ADD",  // 1
"ZEND_SUB",  // 2
"ZEND_MUL",  // 3
"ZEND_DIV",  // 4
"ZEND_MOD",  // 5
"ZEND_SL",  // 6
"ZEND_SR",  // 7
"ZEND_CONCAT",  // 8
"ZEND_BW_OR",  // 9
"ZEND_BW_AND",  // 10
"ZEND_BW_XOR",  // 11
"ZEND_BW_NOT",  // 12
"ZEND_BOOL_NOT",  // 13
"ZEND_BOOL_XOR",  // 14
"ZEND_IS_IDENTICAL",  // 15
"ZEND_IS_NOT_IDENTICAL",  // 16
"ZEND_IS_EQUAL",  // 17
"ZEND_IS_NOT_EQUAL",  // 18
"ZEND_IS_SMALLER",  // 19
"ZEND_IS_SMALLER_OR_EQUAL",  // 20
"ZEND_CAST",  // 21
"ZEND_QM_ASSIGN",  // 22
"ZEND_ASSIGN_ADD",  // 23
"ZEND_ASSIGN_SUB",  // 24
"ZEND_ASSIGN_MUL",  // 25
"ZEND_ASSIGN_DIV",  // 26
"ZEND_ASSIGN_MOD",  // 27
"ZEND_ASSIGN_SL",  // 28
"ZEND_ASSIGN_SR",  // 29
"ZEND_ASSIGN_CONCAT",  // 30
"ZEND_ASSIGN_BW_OR",  // 31
"ZEND_ASSIGN_BW_AND",  // 32
"ZEND_ASSIGN_BW_XOR",  // 33
"ZEND_PRE_INC",  // 34
"ZEND_PRE_DEC",  // 35
"ZEND_POST_INC",  // 36
"ZEND_POST_DEC",  // 37
"ZEND_ASSIGN",  // 38
"ZEND_ASSIGN_REF",  // 39
"ZEND_ECHO",  // 40
"ZEND_PRINT",  // 41
"ZEND_JMP",  // 42
"ZEND_JMPZ",  // 43
"ZEND_JMPNZ",  // 44
"ZEND_JMPZNZ",  // 45
"ZEND_JMPZ_EX",  // 46
"ZEND_JMPNZ_EX",  // 47
"ZEND_CASE",  // 48
"ZEND_SWITCH_FREE",  // 49
"ZEND_BRK",  // 50
"ZEND_CONT",  // 51
"ZEND_BOOL",  // 52
"ZEND_INIT_STRING",  // 53
"ZEND_ADD_CHAR",  // 54
"ZEND_ADD_STRING",  // 55
"ZEND_ADD_VAR",  // 56
"ZEND_BEGIN_SILENCE",  // 57
"ZEND_END_SILENCE",  // 58
"ZEND_INIT_FCALL_BY_NAME",  // 59
"ZEND_DO_FCALL",  // 60
"ZEND_DO_FCALL_BY_NAME",  // 61
"ZEND_RETURN",  // 62
"ZEND_RECV",  // 63
"ZEND_RECV_INIT",  // 64
"ZEND_SEND_VAL",  // 65
"ZEND_SEND_VAR",  // 66
"ZEND_SEND_REF",  // 67
"ZEND_NEW",  // 68
"ZEND_JMP_NO_CTOR",  // 69
"ZEND_FREE",  // 70
"ZEND_INIT_ARRAY",  // 71
"ZEND_ADD_ARRAY_ELEMENT",  // 72
"ZEND_INCLUDE_OR_EVAL",  // 73
"ZEND_UNSET_VAR",  // 74
"ZEND_UNSET_DIM_OBJ",  // 75
"ZEND_ISSET_ISEMPTY",  // 76
"ZEND_FE_RESET",  // 77
"ZEND_FE_FETCH",  // 78
"ZEND_EXIT",  // 79
"ZEND_FETCH_R",  // 80
"ZEND_FETCH_DIM_R",  // 81
"ZEND_FETCH_OBJ_R",  // 82
"ZEND_FETCH_W",  // 83
"ZEND_FETCH_DIM_W",  // 84
"ZEND_FETCH_OBJ_W",  // 85
"ZEND_FETCH_RW",  // 86
"ZEND_FETCH_DIM_RW",  // 87
"ZEND_FETCH_OBJ_RW",  // 88
"ZEND_FETCH_IS",  // 89
"ZEND_FETCH_DIM_IS",  // 90
"ZEND_FETCH_OBJ_IS",  // 91
"ZEND_FETCH_FUNC_ARG",  // 92
"ZEND_FETCH_DIM_FUNC_ARG",  // 93
"ZEND_FETCH_OBJ_FUNC_ARG",  // 94
"ZEND_FETCH_UNSET",  // 95
"ZEND_FETCH_DIM_UNSET",  // 96
"ZEND_FETCH_OBJ_UNSET",  // 97
"ZEND_FETCH_DIM_TMP_VAR",  // 98
"ZEND_FETCH_CONSTANT",  // 99
"ZEND_DECLARE_FUNCTION_OR_CLASS",  // 100
"ZEND_EXT_STMT",  // 101
"ZEND_EXT_FCALL_BEGIN",  // 102
"ZEND_EXT_FCALL_END",  // 103
"ZEND_EXT_NOP",  // 104
"ZEND_TICKS",  // 105
"ZEND_SEND_VAR_NO_REF", // 106
"", // 107
"", // 108
"", // 109
"ZEND_JEXT_START", // 110
"ZEND_JEXT_BOP", // 111
"ZEND_JEXT_FETCH_LOCAL_R_BY_CONST", // 112
"ZEND_JEXT_FETCH_LOCAL_W_BY_CONST", // 113
"ZEND_JEXT_FETCH_LOCAL_RW_BY_CONST", // 114
"ZEND_JEXT_BOP_ASSIGN", // 115
"" // 116
};

template<typename jnode>
string opdist (ex_node& op, const jnode& node, const map<ex_node*,int> p2i)
{
  char buf[1024];

  switch (node.op_type)
    {
      case JOP_FETCH:
      case JOP_CONST:
        {
          const zval &val = *node.u.cnst;
          switch (val.type)
            {
              case IS_NULL:
                snprintf (buf, sizeof (buf), "null");
                break;
              case IS_BOOL:
                snprintf (buf, sizeof (buf), "b(%ld)", val.value.lval);
                break;
              case IS_LONG:
                snprintf (buf, sizeof (buf), "l(%ld)", val.value.lval);
                break;
              case IS_DOUBLE:
                snprintf (buf, sizeof (buf), "d(%f)", val.value.dval);
                break;
              case IS_STRING:
                snprintf (buf, sizeof (buf), "s(\"%s\")", val.value.str.val);
                for ( size_t i = 0; i < 20 && i < strlen (buf); i++ )
                  {
                    if ( buf[i] && !isprint(buf[i]) ) buf[i] = ' ';
                  }
                buf[20] = buf[21] = '.';
                buf[22] = 0;
                break;
              case IS_CONSTANT_ARRAY:
                snprintf (buf, sizeof (buf), "a(..)");
                break;
              default:
                snprintf (buf, sizeof(buf), "ack no way[%d]!", val.type);
                break;
            }
          if ( node.op_type == JOP_FETCH )
              return "fetch{" + string(buf) + "}";
          else
              return string(buf);
          break;
        }
      case JOP_TMP:
        snprintf (buf, sizeof(buf), "tmp[%d].tmp", node.u.var);
        return string(buf);
        break;
      case JOP_VAR:
        snprintf (buf, sizeof(buf), "tmp[%d].var", node.u.var);
        return string(buf);
        break;
      case JOP_UNUSED:
        break;
      case JOP_REG_TYPE:
        snprintf (buf, sizeof (buf), "register(%d)", node.u.register_type);
        return string(buf);
        break;
      case JOP_JMP:
        snprintf (buf, sizeof (buf), "jmp(.j%i)", p2i.find(node.u.jmp_to)->second);
        return string(buf);
        break;
      default:
        snprintf (buf, sizeof (buf), "unknown(%d)", node.op_type);
        return string(buf);
        break;
    }

  return "";
}

char * jphp_disassemble (zend_op_array *op_array)
{
  if ( !op_array || op_array->type != ZEND_USER_FUNCTION )
    {
      zend_error (E_WARNING, "JIT-PHP: disassemble: no op_array found or this is not a user defined function!");
      return NULL;
    }

  if ( !op_array->opcodes )
    {
      zend_error (E_WARNING, "JIT-PHP: disassemble: no opcodes found for given op_array!");
      return NULL;
    }

  if ( op_array->opcodes[0].opcode != ZEND_JEXT_START )
    {
      zend_error (E_WARNING, "JIT-PHP: disassemble: not a jphp encoded script!");
      return NULL;
    }

  ex_node *opline = (ex_node *)op_array->opcodes;

  string ret;

  map<ex_node *,int> p2i; // ptr2index
  set<ex_node *> j2;      // jump2

  typedef map<int (*)(zval *, zval *, zval *), const char *> bop2name_t;
  bop2name_t bop2name;
  bop2name[add_function] = "add";
  bop2name[sub_function] = "sub";
  bop2name[mul_function] = "mul";
  bop2name[div_function] = "div";
  bop2name[shift_left_function] = "<<";
  bop2name[shift_right_function] = ">>";
  bop2name[concat_function] = "concat";
  bop2name[bitwise_or_function] = "bit_or";
  bop2name[bitwise_and_function] = "bit_and";
  bop2name[bitwise_xor_function] = "bit_xor";
  bop2name[is_identical_function] = "===";
  bop2name[is_not_identical_function] = "!==";
  bop2name[is_equal_function] = "==";
  bop2name[is_not_equal_function] = "!=";
  bop2name[is_smaller_function] = "<";
  bop2name[is_smaller_or_equal_function] = "<=";
  bop2name[boolean_xor_function] = "bool_xor";

  for ( size_t i = 0; i < op_array->size; i++ )
    {
      ex_node &op = opline[i];
      p2i[&op] = i;

      if ( op.jext_type == JEXT_ALT_JMP )
          j2.insert (op.jex.alt_jump.next);
      if ( op.op2.op_type == JOP_JMP )
          j2.insert (op.op2.u.jmp_to);
    }

  for ( zend_uint i = 0; i < op_array->size; i++ )
    {
      ex_node &op = opline[i];
      string line;

      if ( j2.find(&op) != j2.end() )
        {
          char buf[16];
          snprintf (buf, sizeof (buf), ".j%d:\n", i);
          ret += buf;
        }

      line = "    ";

      string on = opname[op.opcode];
      on.erase (on.begin(), on.begin()+5);
      on.erase (remove (on.begin(), on.end(), '_'), on.end());
      transform (on.begin(), on.end(), on.begin(), static_cast<int(*)(int)> (std::tolower));

      line += on;

      for ( size_t j = 0; j < 25 - on.size(); j++ )
          line.push_back (' ');

      string arg = "";

      string a = opdist (op, op.op1, p2i);
      arg += a;

      a = opdist (op, op.op2, p2i);
      if ( a != "" )
        {
          if ( arg != "" ) arg += ", ";
          arg += a;
        }

      a = opdist (op, op.result, p2i);
      if ( a != "" )
        {
          arg += " => ";
          arg += a;
          if ( op.result.EA_type&EXT_TYPE_UNUSED ) arg += ":unused";
        }

      line += arg;

      string jta;
      if ( arg != "" ) jta += " ";

      switch (op.jext_type)
        {
          case JEXT_UNUSED:
            break;
          case JEXT_ALT_JMP:
            {
              char buf[32];
              snprintf (buf, sizeof(buf), "{jmp(.j%d)}", p2i.find(op.jex.alt_jump.next)->second);
              jta += buf;
            }
            break;
          case JEXT_BOP:
            {
              char buf[64];
              bop2name_t::iterator it = bop2name.find(op.jex.bop.op);
              if ( it == bop2name.end() )
                  snprintf (buf, sizeof(buf), "{bop:unknown}");
              else
                  snprintf (buf, sizeof(buf), "{bop:%s}", it->second);
              jta += buf;
            }
            break;
          case JEXT_FENTRY:
            jta += "{fentry}";
            break;
          case JEXT_CENTRY:
            jta += "{centry}";
            break;
          case JEXT_HEADER:
            jta += "{header}";
            break;
          default:
            jta += "{unknown}";
            break;
        }

      if ( jta != "" && jta != " " ) line += jta;

      line += "\n";

      ret += line;
    }

  return estrdup (ret.c_str());
}


