
//
// JPHP - JIT optimizer for the PHP language
// Copyright (C) 2001-2003 Indrek Mandre <indrek(at)mare.ee>
//
// This source code is made available free and without charge subject
// to the terms of the QPL as detailed in bundled LICENSE.QPL file.
//
// It may be only used for non-commercial purpose.
//

#ifndef __execute_h__
#define __execute_h__

#ifdef VALGRIND_DEBUG
#include <valgrind/memcheck.h>
#endif

#define ZEND_JEXT_START 110
#define ZEND_JEXT_BOP 111
#define ZEND_JEXT_FETCH_LOCAL_R_BY_CONST 112
#define ZEND_JEXT_FETCH_LOCAL_W_BY_CONST 113
#define ZEND_JEXT_FETCH_LOCAL_RW_BY_CONST 114
#define ZEND_JEXT_BOP_ASSIGN 115

#define JEXT_UNUSED 0
#define JEXT_ALT_JMP 1
#define JEXT_BOP 2
#define JEXT_FENTRY 3
#define JEXT_HEADER 4
#define JEXT_CENTRY 5

#define JOP_UNUSED 0
#define JOP_CONST 1
#define JOP_TMP 2
#define JOP_VAR 3
#define JOP_FETCH 4
#define JOP_JMP 5
#define JOP_REG_TYPE 6

typedef struct ex_node ex_node;
typedef struct jphp_fentry jphp_fentry;
typedef struct jphp_header jphp_header;
typedef struct jphp_centry jphp_centry;
typedef struct jphp_const jphp_const;

struct jphp_const {
  zval val;
  uint hcode;
};

struct jphp_fentry
{
  zend_op_array *op_array;
  zend_uchar bind_at_load;
  char *name;
  uint name_len;
  uint name_hcode;
};

struct jphp_centry
{
  zend_class_entry *centry;
  zend_uchar bind_at_load;

  uint function_count;

  char *name;
  uint name_len;
  uint name_hcode;

  char *parent;
  uint parent_len;
  uint parent_hcode;

  HashTable *function_table;
  HashTable *default_properties;
};

struct jphp_header
{
  char *path;

  uint function_count;
  jphp_fentry *functions;

  uint class_count;
  jphp_centry *classes;

  uint const_count;
  jphp_const *consts;

  time_t stmtime;

  char *shared_path;
  uint is_shared;
  uint user_count; 
  uint usage_count;
  time_t script_last_accessed;

  jphp_header *next;
  jphp_header *prev;

  jphp_header *h_next;
  jphp_header *h_prev;
};

typedef struct _jnode {
  zend_uchar op_type;
  zend_uchar EA_type;
  union {
    zval *cnst;
    jphp_const *jconst;
    ex_node *jmp_to;

    zend_uint var;
    zend_uint opline_num;
    zend_uint fetch_type;
    zend_uint register_type;
  } u;
} jnode;

struct ex_node
{
  zend_uchar opcode;
  zend_uchar jext_type;

  jnode result;
  jnode op1;
  jnode op2;

  uint lineno;

  union {
    struct {
      ex_node *next;
    } alt_jump;

    struct {
        int (*op)(zval *, zval *, zval *);
    } bop;

    jphp_fentry *fentry;
    jphp_centry *centry;
    jphp_header *header;

    ulong extended_value;
  } jex;

  void *jump;
};

#ifdef __cplusplus
extern "C" {
#endif

extern void (*old_zend_execute)(zend_op_array *op_array TSRMLS_DC);
extern void jphp_execute(zend_op_array *op_array);
extern void *jphp_jump[128];

extern void jphp_register_function (ex_node *node);
extern int jphp_function_lineno (zend_function *function);

#ifdef __cplusplus
};
#endif

#endif // __execute_h__

