
//
// JPHP - JIT optimizer for the PHP language
// Copyright (C) 2001-2003 Indrek Mandre <indrek(at)mare.ee>
//
// This source code is made available free and without charge subject
// to the terms of the QPL as detailed in bundled LICENSE.QPL file.
//
// It may be only used for non-commercial purpose.
//

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <sys/mman.h>

#include <limits.h>
#include <unistd.h>
#include <string.h>

#include "freplace.hh"

extern "C" int getpagesize ();

int _function_replace (struct freplace_data* store_old, void *f1, void *f2)
{
  unsigned long a1;
  unsigned long a2;
  unsigned long diff;
  unsigned char *mptr;
  unsigned char *ptr = (unsigned char *)f1 - ((unsigned int)f1) % getpagesize();

  if ( store_old ) store_old->function = NULL;

  if ( mprotect (ptr, 4096, PROT_WRITE|PROT_READ|PROT_EXEC) )
      return 0;

  if ( store_old )
    {
      store_old->function = f1;
      memcpy (store_old->head, f1, sizeof (store_old->head));
    }

  a1 = (unsigned long)f1;
  a2 = (unsigned long)f2;

  mptr = (unsigned char *)f1;

#ifdef sparc
  if ( a2 > a1 )
    {
      diff = a2 - a1;
    }
  else
    {
      diff = 0x1000000 - (a1 - a2);
    }

  assert (diff >= 0x1000000);

  diff >>= 2; /* jumps are 4 byte aligned */
  diff |= 0x10800000; /* add missing jump operand bits */
  memcpy (mptr, &diff, 4);
  mptr += 4;

  *mptr++ = 0x01; /* nop */
  *mptr++ = 0x00;
  *mptr++ = 0x00;
  *mptr++ = 0x00;
#else /* x386 */
  *mptr = 0xe9;  // long jmp, relative to position
  mptr++;
  diff = a2 - a1;
  diff -= 5;    // long jump size
  memcpy (mptr, &diff, 4);
#endif

  mprotect (ptr, 4096, PROT_READ|PROT_EXEC);

  return 1;
}

int _function_restore (struct freplace_data* old)
{
  void *f1 = old->function;
  unsigned char *ptr = (unsigned char *)f1 - ((unsigned int)f1) % getpagesize();

  if ( !old->function ) return 0;

  if ( mprotect (ptr, 4096, PROT_WRITE|PROT_READ|PROT_EXEC) )
      return 0;

  memcpy (f1, old->head, sizeof (old->head));

  mprotect (ptr, 4096, PROT_READ|PROT_EXEC);

  return 1;
}

