
//
// JPHP - JIT optimizer for the PHP language
// Copyright (C) 2001-2003 Indrek Mandre <indrek(at)mare.ee>
//
// This source code is made available free and without charge subject
// to the terms of the QPL as detailed in bundled LICENSE.QPL file.
//
// It may be only used for non-commercial purpose.
//

#include "jphp.hh"

zend_op_array * Jphp::compile (zend_file_handle *file_handle)
{
  char *fn = file_handle->opened_path ? file_handle->opened_path : file_handle->filename;
  if ( !fn ) return NULL;

  vstring path(fn);

  jphp_header *hdr;
  rs_t::iterator it = runtime_scripts.find (path);

  if ( it == runtime_scripts.end() )
    {
      hdr = JphpSystem::aquire_script (file_handle);
      if (hdr) runtime_scripts[path] = hdr;
    }
  else
    {
      hdr = it->second;
      zend_llist_add_element(&CG(open_files), file_handle);
    }

  if ( !hdr ) return NULL;

  // now register functions and classes

  for ( size_t i = 1; i < hdr->function_count; i++ )
    {
      if ( !hdr->functions[i].bind_at_load ) break;

      jphp_fentry& fentry = hdr->functions[i];

      zend_function *function, *rfun;
      php_hash<zend_function> ft(EG(function_table));

      function = reinterpret_cast<zend_function *>(fentry.op_array);

      if ( !ft.quick_insert (fentry.name, fentry.name_len + 1, fentry.name_hcode, *function, rfun) ) {
        if ( ft.quick_find (fentry.name, fentry.name_len + 1, fentry.name_hcode, function)
          && function->type==ZEND_USER_FUNCTION
          && ((zend_op_array *) function)->last>0) {
          zend_error(E_ERROR, "Cannot redeclare %s() (previously declared in %s:%d)",
                fentry.name,
                ((zend_op_array *) function)->filename,
                jphp_function_lineno (function));
        } else {
          zend_error(E_ERROR, "Cannot redeclare %s()", fentry.name);
        }
      } else {
        if ( rfun->op_array.static_variables )
          {
            HashTable *ht = rfun->op_array.static_variables;
            zval *tmp;

            ALLOC_HASHTABLE_REL(rfun->op_array.static_variables);
            zend_hash_init(rfun->op_array.static_variables, 0, NULL, ZVAL_PTR_DTOR, 0);
            zend_hash_copy(rfun->op_array.static_variables, ht, (copy_ctor_func_t) zval_add_ref,
                (void *) &tmp, sizeof(zval *));
          }

        static zend_uint large_refcount;
        large_refcount  = 666666666;
        rfun->op_array.refcount = &large_refcount;
      }
    }

  for ( size_t i = 0; i < hdr->class_count; i++ )
    {
      if ( !hdr->classes[i].bind_at_load ) break;

      jphp_centry& centry = hdr->classes[i];

      jphp_register_class (centry);
    }

  return hdr->functions[0].op_array;
}

void Jphp::start_request ()
{
}

void Jphp::end_request ()
{
  for ( rs_t::iterator it = runtime_scripts.begin(); it != runtime_scripts.end(); it++ )
      JphpSystem::release_script (it->second);

  runtime_scripts.clear();
}


