
//
// JPHP - JIT optimizer for the PHP language
// Copyright (C) 2001-2003 Indrek Mandre <indrek(at)mare.ee>
//
// This source code is made available free and without charge subject
// to the terms of the QPL as detailed in bundled LICENSE.QPL file.
//
// It may be only used for non-commercial purpose.
//

#ifndef __jphp_hh__
#define __jphp_hh__

#include <cstdio>
#include <cstdlib>
#include <cassert>
#include <cctype>

#include <string>
#include <vector>
#include <set>
#include <ext/hash_set> // hash_set for older gcc
#include <map>
#include <ext/hash_map> // hash_map for older gcc
#include <list>
#include <iostream>
#include <fstream>

#include <functional>
#include <algorithm>
#include <numeric>

using namespace std;  // I know, evil
using namespace __gnu_cxx; // delete this for older gcc

#include <sys/types.h>
#include <sys/stat.h>

#include "php_runtime.h"
#include "execute.h"

#include "data.hh"
#include "utils.hh"

extern void jphp_optimize (const zend_op_array& op_array,
    const map<vstring, int>& f2index, Dataoutput& data, ObjectCache<jphp_const>& consts, ostream& out);

extern int jphp_encode (const char *srcpath, const char *targetpath, bool print_header = true);
extern int jphp_encode (zend_file_handle *file_handle, ostream& out, bool print_header);

extern struct jphp_header * jphp_load (char *buf, size_t buf_len);

extern zend_op_array * jphp_compile (zend_file_handle *file_handle, int type);

extern char * jphp_disassemble (zend_op_array *op_array);
extern void jphp_register_class (jphp_centry& centry);

extern void jphp_alloc_init ();

class Lock;
class SharedMem;

class Jphp {
public:
  void start_request ();
  zend_op_array * compile (zend_file_handle *file_handle);
  void end_request ();

private:
  typedef hash_map<vstring, jphp_header *, vstring_hash_func> rs_t;
  rs_t runtime_scripts;
};


namespace JphpSystem {
  Jphp * get_running_context ();

  jphp_header * aquire_script (zend_file_handle *file_handle);
  void release_script (jphp_header *script);

  void init ();
};

#endif // __jphp_hh__

