
//
// JPHP - JIT optimizer for the PHP language
// Copyright (C) 2001-2003 Indrek Mandre <indrek(at)mare.ee>
//
// This source code is made available free and without charge subject
// to the terms of the QPL as detailed in bundled LICENSE.QPL file.
//
// It may be only used for non-commercial purpose.
//

#include <cstdio>
#include <ctime>

#include <string>
#include <vector>
#include <set>
#include <map>
#include <fstream>

#include <sys/param.h>
#include <unistd.h>
#include <signal.h>
#include <sys/time.h>

#include "php_runtime.h"
#include "execute.h"
#include "jphp.hh"
#include "execute.h"
#include "freplace.hh"

ZEND_INI_BEGIN()
  ZEND_INI_ENTRY("jit.enable",                        "1",   0,   NULL)
  ZEND_INI_ENTRY("jit.data_dir",                      "/tmp/jphp",   0,   NULL)

  ZEND_INI_ENTRY("jit.register_development",          "0",   0,   NULL)

  ZEND_INI_ENTRY("jit.enable_file_cache",             "1",   ZEND_INI_SYSTEM|ZEND_INI_PERDIR,   NULL)
  ZEND_INI_ENTRY("jit.enable_shared_memory_cache",    "1",   ZEND_INI_SYSTEM|ZEND_INI_PERDIR,   NULL)

  ZEND_INI_ENTRY("jit.file_perm",                     "444",       0,   NULL)
  ZEND_INI_ENTRY("jit.dir_perm",                      "777",       0,   NULL)
  ZEND_INI_ENTRY("jit.lock_perm",                     "666",       0,   NULL)

  ZEND_INI_ENTRY("jit.shared_memory_perm",            "0666",       0,   NULL)
  ZEND_INI_ENTRY("jit.shared_memory_size",            "16777216",   0,   NULL)
  ZEND_INI_ENTRY("jit.shared_memory_key",             "0xc0dec00",  0,   NULL)
  ZEND_INI_ENTRY("jit.shared_memory_destroy_at_exit", "1",          0,   NULL)
  ZEND_INI_ENTRY("jit.shared_memory_hash_size",       "1024",       0,   NULL)
  ZEND_INI_ENTRY("jit.shared_memory_max_cached_script_size", "2000000",    0,   NULL)
  ZEND_INI_ENTRY("jit.shared_memory_max_cached_script_ttl",  "21600",      0,   NULL)
ZEND_INI_END()

static ZEND_FUNCTION(jdisassemble);
static ZEND_FUNCTION(jencode);

static zend_function_entry jphp_functions[] = {
  ZEND_FE(jdisassemble, NULL)
  {NULL, NULL, NULL}
};

extern "C" char *expand_filepath(const char *filepath, char *real_path TSRMLS_DC);

void (*old_zend_execute)(zend_op_array *op_array TSRMLS_DC);
static zend_op_array *(*old_zend_compile_file)(zend_file_handle *file_handle, int type);

extern "C" int cfg_get_string(char *varname, char **result);

static zend_op_array *jphp_zend_compile_file (zend_file_handle *file_handle, int type)
{
  char *val;

  if ( !file_handle->filename ||  // hmm, can this happen?
      strcmp (file_handle->filename, "-") == 0 || // stdio, anything else?
      strchr (file_handle->filename, ':') ||  // streams of any kind, protocols
      file_handle->type == ZEND_HANDLE_FP && file_handle->handle.fp == stdin || // stdin to be sure
      !INI_BOOL ("jit.enable_file_cache") )
    {
      return old_zend_compile_file (file_handle, type);
    }

  if ( file_handle->type == ZEND_HANDLE_FILENAME &&
      file_handle->filename[0] == '/' &&
      !strstr (file_handle->filename, "..") &&
      !INI_BOOL ("safe_mode") )
    {
      file_handle->handle.fp = fopen (file_handle->filename, "r");
      file_handle->type = ZEND_HANDLE_FP;
    }

  switch (file_handle->type) {
    case ZEND_HANDLE_FILENAME:
      file_handle->handle.fp = zend_fopen(file_handle->filename, &file_handle->opened_path);
      break;
    case ZEND_HANDLE_FD:
      file_handle->handle.fp = fdopen(file_handle->handle.fd, "r");
      break;
    case ZEND_HANDLE_FP:
      break;
  }

  file_handle->type = ZEND_HANDLE_FP;

  if ( !file_handle->handle.fp )
    {
      if (type==ZEND_REQUIRE) {
        zend_message_dispatcher(ZMSG_FAILED_REQUIRE_FOPEN, file_handle->filename);
        zend_bailout();
      } else {
        zend_message_dispatcher(ZMSG_FAILED_INCLUDE_FOPEN, file_handle->filename);
      }

      return NULL;
    }

  HANDLE_BLOCK_INTERRUPTIONS ();  // to be on the safe side
  typedef void (*my_sighandler_t)(int);
  my_sighandler_t old_alarm = signal (SIGALRM, SIG_IGN);  // sorry about this but we can't let any1 interrupt
  my_sighandler_t old_prof = signal (SIGPROF, SIG_IGN);  // sorry about this but we can't let any1 interrupt
  struct itimerval time_left;
  getitimer (ITIMER_PROF, &time_left);
  zend_op_array *op_array = JphpSystem::get_running_context()->compile (file_handle);
  setitimer (ITIMER_PROF, &time_left, NULL);
  signal (SIGPROF, old_prof);
  signal (SIGALRM, old_alarm);
  HANDLE_UNBLOCK_INTERRUPTIONS ();

  if ( !op_array )
    {
      if (type==ZEND_REQUIRE) {
        zend_message_dispatcher(ZMSG_FAILED_REQUIRE_FOPEN, file_handle->filename);
        zend_bailout();
      } else {
        zend_message_dispatcher(ZMSG_FAILED_INCLUDE_FOPEN, file_handle->filename);
      }

      return NULL;
    }

  zend_op_array *ret_op_array = static_cast<zend_op_array *>(emalloc (sizeof (zend_op_array)));
  memcpy (ret_op_array, op_array, sizeof (zend_op_array));

  static zend_uint large_refcount;
  large_refcount = 666666666;
  ret_op_array->refcount = &large_refcount;

  if ( ret_op_array->static_variables )
    {
      HashTable *ht = ret_op_array->static_variables;
      zval *tmp;

      ALLOC_HASHTABLE_REL(ret_op_array->static_variables);
      zend_hash_init(ret_op_array->static_variables, 0, NULL, ZVAL_PTR_DTOR, 0);
      zend_hash_copy(ret_op_array->static_variables, ht, (copy_ctor_func_t) zval_add_ref,
          (void *) &tmp, sizeof(zval *));
    }

  return ret_op_array;
}

static uint my_zend_get_executed_lineno(TSRMLS_D)
{
  zend_execute_data *data = EG(current_execute_data);

  if ( data && data->op_array && data->op_array->opcodes && data->op_array->opcodes[0].opcode == ZEND_JEXT_START )
      return ((ex_node*)active_opline)->lineno;

  if (EG(opline_ptr))
      return active_opline->lineno;

  return 0;
}

ZEND_FUNCTION(jdisassemble)
{
  zval **name;
  zend_op_array *array;
  char *ret;

  if ( ZEND_NUM_ARGS() > 1 )
      ZEND_WRONG_PARAM_COUNT() 

  array = EG(active_op_array);

  if ( ZEND_NUM_ARGS() == 1 )
    {
      zend_get_parameters_ex (1, &name);
      convert_to_writable_string_ex (name); // should this be destroyed later?
      char *str = Z_STRVAL_PP(name);
      zend_str_tolower (str, Z_STRLEN_PP(name));
      char *class_sep = strstr (str, "::");

      if ( class_sep )
        {
          char *fn = class_sep + 2;
          *class_sep = 0;
          php_hash<zend_class_entry> ct(EG(class_table));
          zend_class_entry *ce = ct.find (str);
          if ( !ce )
            {
              zend_error (E_WARNING, "JIT-PHP: Unknown class (%s), unable to disassemble!", str);
              RETURN_FALSE;
            }

          php_hash<zend_function> ft(&ce->function_table);
          zend_function *f = ft.find (fn);

          if ( !f )
            {
              zend_error (E_WARNING, "JIT-PHP: Unknown class::method (%s::%s), unable to disassemble!", str, fn);
              RETURN_FALSE;
            }

          array = &f->op_array;
        }
      else
        {
          php_hash<zend_function> ft(EG(function_table));
          zend_function *f = ft.find (str);

          if ( !f )
            {
              zend_error (E_WARNING, "JIT-PHP: Unknown function (%s), unable to disassemble!", str);
              RETURN_FALSE;
            }

          array = &f->op_array;
        }
    }

  ret = jphp_disassemble (array);

  if ( !ret )
      RETURN_FALSE;

  RETURN_STRING(ret, 0);
}

static int jphp_ext_startup (zend_extension *extension)
{
  {
    int module_number = 0;
    REGISTER_INI_ENTRIES();
  }

  JphpSystem::init ();

  if ( !INI_BOOL ("jit.enable") ) return SUCCESS;

  if ( INI_BOOL ("jit.register_development") )
      zend_register_functions(jphp_functions, NULL, MODULE_PERSISTENT TSRMLS_CC);

  jphp_execute (NULL);  // init the jump table

  old_zend_execute = zend_execute;  // override execute
  zend_execute = jphp_execute;

  old_zend_compile_file = zend_compile_file;  // and compile function pointers
  zend_compile_file = jphp_zend_compile_file;

  freplace (NULL, zend_get_executed_lineno, my_zend_get_executed_lineno);

  return SUCCESS;
}

static void jphp_ext_shutdown (zend_extension *extension)
{
}

static void jphp_ext_activate ()
{
  JphpSystem::get_running_context()->start_request ();
}

static void jphp_ext_deactivate ()
{
  JphpSystem::get_running_context()->end_request ();
}

extern "C" {

zend_extension zend_extension_entry = {
  "JIT PHP optimizer",
  "0.1",
  "Indrek Mandre",
  "http://www.mare.ee/indrek/jphp/",
  "Copyright (c) 2001-2003 Indrek Mandre",

  jphp_ext_startup,           // startup
  jphp_ext_shutdown,          // shutdown
  jphp_ext_activate,          // activate
  jphp_ext_deactivate,        // deactivate

  NULL,                       // message_handler, only have ZEND_EXTMSG_NEW_EXTENSION there now
  NULL,                       // op_array_handler, some sort of operands optimizer

  NULL,                       // statement_handler, always called on operand ZEND_EXT_STMT, in zend_execute
  NULL,                       // fcall_begin_handler, always called on operand ZEND_EXT_FCALL_BEGIN, in zend_execute
  NULL,                       // fcall_end_handler, always called on operand ZEND_EXT_FCALL_END, in zend_execute

  NULL,                       // op_array_ctor, in zend_opcode, preprocess opcode array before compilation
  NULL,                       // op_array_dtor, in zend_opcode, postprocesses opcodes after use

  NULL,                       // api_check_no, let it be, we are not dependant on API-s

  NULL,                       // reserved2
  NULL,                       // reserved3
  NULL,                       // reserved4
  NULL,                       // reserved5
  NULL,                       // reserved6
  NULL,                       // reserved7
  NULL,                       // reserved8

  NULL,                       // handle?
  0                           // resource number?
};


zend_extension_version_info extension_version_info = {
  ZEND_EXTENSION_API_NO,      // zend_extension_api_no
  "4.3.0",                    // required_zend_version
  0,                          // thread_safe
  0,                          // debug
};

}; // extern "C"

