
//
// JPHP - JIT optimizer for the PHP language
// Copyright (C) 2001-2003 Indrek Mandre <indrek(at)mare.ee>
//
// This source code is made available free and without charge subject
// to the terms of the QPL as detailed in bundled LICENSE.QPL file.
//
// It may be only used for non-commercial purpose.
//

#include "jphp.hh"
#include "mm.hh"
#include "shared.hh"

#include <unistd.h>
#include <sys/file.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/shm.h>
#include <fcntl.h>

Lock::~Lock ()
{
}

FileLock::FileLock (mode_t m) : fd(-1), locked (LOCK_UN), mode(m)
{
}

FileLock::FileLock (const vstring& name, mode_t m) : fd(-1), locked (LOCK_UN), mode(m)
{
  init (name);
}

bool FileLock::init (const vstring& name)
{
  if ( fd != -1 ) return false;

  path = name;
  mode_t m = umask(0);
  fd = open (name.c_str(), O_CREAT|O_RDWR, mode);
  umask (m);

  if ( fd == -1 ) return false;

  return true;
}

void FileLock::read ()
{
  if ( locked == LOCK_SH ) return;
  if ( locked != LOCK_UN ) unlock ();
  flock (fd, LOCK_SH);
  locked = LOCK_SH;
}

void FileLock::lock ()
{
  if ( locked == LOCK_EX ) return;
  if ( locked != LOCK_UN ) unlock ();
  flock (fd, LOCK_EX);
  locked = LOCK_EX;
}

void FileLock::unlock ()
{
  if ( locked == LOCK_UN ) return;
  flock (fd, LOCK_UN);
  locked = LOCK_UN;
}

FileLock::~FileLock ()
{
  if ( locked != LOCK_UN ) unlock ();
  close (fd);
}

bool DummyLock::init (const vstring& name) { return true; }
void DummyLock::read () { }
void DummyLock::lock () { }
void DummyLock::unlock () { }
DummyLock::~DummyLock () { }


SharedMem::SharedMem ()
{
}

SharedMem::~SharedMem ()
{
}

IPCSharedMem::IPCSharedMem () : mem_size (0), mem(NULL)
{
}

SharedMem::InitStatus IPCSharedMem::init (void *base_ptr, const vstring& name, size_t size)
{
  InitStatus ret = attached;

  if ( this->mem_size != 0 || size == 0 ) return error;

  if ( strncmp ("0x", name.c_str(), 2) == 0 )
      key = strtol (name.c_str(), NULL, 16);
  else
      key = atoi (name.c_str());

  if ( !key )
      key = 0xc0dec00;

  if ( !key )
      return error;

  shmid = shmget (key, size, IPC_CREAT|IPC_EXCL|strtol(INI_STR("jit.shared_memory_perm"), NULL, 8));
  if ( shmid == -1 && errno == EEXIST )
      shmid = shmget (key, size, strtol(INI_STR("jit.shared_memory_perm"), NULL, 8));
  else
      ret = created;

  if ( shmid == -1 )
      return error;

  mem = shmat (shmid, base_ptr, 0);

  if ( !mem )
      return error;

  this->mem_name = name;
  this->mem_size = size;

  return ret;
}

void IPCSharedMem::delete_at_exit ()
{
  if ( mem )
      shmctl (shmid, IPC_RMID, NULL);
}

void * IPCSharedMem::get_mem ()
{
  return mem;
}

size_t IPCSharedMem::size () const
{
  return mem_size;
}

const vstring& IPCSharedMem::get_name () const
{
  return mem_name;
}

void IPCSharedMem::detach ()
{
  if ( !mem_size ) return;
  mem_size = 0;
  shmdt (mem);
  mem = NULL;
}

IPCSharedMem::~IPCSharedMem ()
{
}


