
//
// JPHP - JIT optimizer for the PHP language
// Copyright (C) 2001-2003 Indrek Mandre <indrek(at)mare.ee>
//
// This source code is made available free and without charge subject
// to the terms of the QPL as detailed in bundled LICENSE.QPL file.
//
// It may be only used for non-commercial purpose.
//

#ifndef __shared_hh__
#define __shared_hh__

class Lock
{
public:
  virtual bool init (const vstring& name) = 0;

  virtual void read () = 0;
  virtual void lock () = 0;
  virtual void unlock () = 0;

  virtual ~Lock ();
};

class DummyLock : public Lock
{
public:
  virtual bool init (const vstring& name);

  virtual void read ();
  virtual void lock ();
  virtual void unlock ();

  virtual ~DummyLock ();
};

class FileLock : public Lock
{
public:
  FileLock (mode_t mode = 0600);
  FileLock (const vstring& name, mode_t mode = 0600);

  virtual bool init (const vstring& name);

  virtual void read ();
  virtual void lock ();
  virtual void unlock ();

  virtual ~FileLock ();

private:
  vstring path;
  int fd;
  int locked;
  mode_t mode;
};

class MemMan;

class SharedMem
{
public:
  SharedMem ();
  virtual ~SharedMem ();

  enum InitStatus { error = -1, attached = 0, created = 1 };

  virtual InitStatus init (void *base_ptr, const vstring& name, size_t size) = 0;
  virtual void delete_at_exit () = 0;
  virtual void * get_mem () = 0;
  virtual void detach () = 0;
  virtual size_t size () const = 0;
  virtual const vstring& get_name () const = 0;
};

class IPCSharedMem : public SharedMem
{
public:
  IPCSharedMem ();

  virtual InitStatus init (void *base_ptr, const vstring& name, size_t size);
  virtual void delete_at_exit ();
  virtual void * get_mem ();
  virtual void detach ();
  virtual size_t size () const;
  virtual const vstring& get_name () const;

  virtual ~IPCSharedMem ();

private:
  unsigned long key;
  int shmid;
  vstring mem_name;
  size_t mem_size;
  void *mem;
};

#endif // !__shared_hh__

